"""utility function for interacting with waze api"""

import json
import datetime as dt

import requests

from waze_logger.location import Location


WAZE_URL = "https://waze.com"
ROUTING_ENDPOINT = "RoutingManager/routingRequest"

DEFAULT_HEADERS = {
    "User-Agent": "PostmanRuntime/7.33.0"
}


def get_route(origin: Location, destination: Location) -> dict:

    params = {
        "from": str(origin),
        "to": str(destination),
        "returnJSON": "true",
        "at": 0,
        "nPaths": 1,
    }

    url = f"{WAZE_URL}/{ROUTING_ENDPOINT}"

    response = requests.get(url=url, params=params, headers=DEFAULT_HEADERS)

    return json.loads(response.text)["response"]


def get_route_duration(
    origin: Location,
    destination: Location,
) -> int:

    route = get_route(origin, destination)

    return route["totalRouteTime"]


if __name__ == '__main__':

    ORIGIN = Location(45.760919431034004, -73.98788574519124)
    DESTINATION = Location(45.47939592085909, -73.805993034125)

    location_str = str(ORIGIN)

    route = get_route_duration(ORIGIN, DESTINATION)

    foo = "bar"



"""module containing utility function for sqlachemy engine"""

from urllib.parse import quote as url_encode

from sqlalchemy import Engine
import sqlalchemy as db


def create_engine_url(
    engine: str,
    username: str,
    password: str,
    host: str,
    port: int = None,
    database: str = None,
    query: dict = None,
    connector: str = None,
) -> str:
    """creates an sql engine url based on the parameter provided"""

    engine_url = engine

    if connector is not None:
        engine_url += f"+{connector}"

    engine_url += f"://{username}:{password}@{host}"

    if port is not None:
        engine_url += f":{port}"

    if database is not None:
        engine_url += f"/{database}"

    if query is not None:
        engine_url += f"?{x_url_encode_form(query)}"

    return engine_url


def create_engine(credentials: dict) -> Engine:
    """creates an engine from the credentials provided. Check
    'create_engine_url' for the keys required in the credential
    dictionary"""

    engine_url = create_engine_url(**credentials)

    return db.create_engine(engine_url)


def x_url_encode_form(query: dict):
    """url encode a dictionary for an url based query"""

    items = []

    for key, value in query.items():
        key = url_encode(key)
        value = url_encode(value)

        items.append(f"{key}={value}")

    return "&".join(items)



"""main module containing the cli"""

import argparse
import logging
import json
import csv
import datetime as dt
from time import sleep
from pathlib import Path

from sqlalchemy import Engine, Table, MetaData

from waze_logger.database.engine import create_engine
from waze_logger.database.create import create_database
from waze_logger.location import Location
from waze_logger.route import Route

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def get_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(
        "waze_logger",
        description=(
            "A utility to log waze route time from their API into a database"
        ),
    )

    parser.add_argument(
        "-c", "--config-file",
        help=(
            "location of the configuration file containing the credentials for"
            " the database"
        ),
        default=Path.home() / ".config/waze_logger/config.json",
        type=Path,
    )

    parser.add_argument(
        "-s", "--setup",
        help="creates the setup make to create the database.",
        default=False,
    )

    parser.add_argument(
        "-f", "--force",
        help="forces the setup of the database even if it contains data",
        default=False,
    )

    parser.add_argument(
        "--locations",
        help="path to a csv file containing location to include in the setup",
        default=None,
        type=Path,
    )

    return parser.parse_args()


def main():

    args = get_args()

    LOGGER.info("Reading the configuration file")

    with open(args.config_file, 'r', encoding="utf8") as file:
        config = json.load(file)

    engine = create_engine(config["database"])

    if args.setup:
        LOGGER.info("Setting the database")
        locations = _prep_location(args.locations)
        create_database(engine, locations, args.force)
        return None

    # get the list of routes from the database
    routes = Route.from_database(engine)

    log_routes(routes, engine)


def log_routes(routes: list[Route], engine: Engine):

    if len(routes) == 0:
        return None

    durations = []

    for route in routes:

        duration = route.duration()

        durations.append({
            "route": route.id,
            "trip_duration": duration,
            "logged_time": dt.datetime.now(),
        })
        LOGGER.info(f"route {route.id} -> {duration}")
        sleep(1)

    with engine.connect() as connection:

        table = Table("route_duration", MetaData(), autoload_with=engine)
        connection.execute(table.insert(), durations)
        connection.commit()


def _prep_location(file_path: str) -> list[Location]:

    if file_path is None:
        return None

    with open(file_path, 'r', encoding="utf8") as file:
        csv_file = csv.reader(file)

        headers = next(csv_file)

        locations = []

        for row in csv_file:

            record = dict(zip(headers, row))
            locations.append(Location(**record))

    return locations


if __name__ == '__main__':

    try:
        main()
    except KeyboardInterrupt:
        LOGGER.info("Aborting process")



"""module used for the creation of the database on setup"""

from itertools import combinations

import sqlalchemy as db

from waze_logger.location import Location
from waze_logger.route import Route


def create_database(
        engine: db.Engine,
        locations: list[Location] = None,
        force: bool = False,
):

    # check if tables exist and if so drop before continuing
    metadata = db.MetaData()
    metadata.reflect(engine)

    if len(metadata.tables.keys()) > 0 and not force:
        raise ValueError(
            "The database contains tables. Aborting. To clear the database,"
            "set 'force' to true"
        )
    elif len(metadata.tables.keys()) > 0 and force:
        metadata.drop_all(bind=engine)

    _create_table(engine)

    if locations is not None:

        # create the locations in the database and retrieve them with
        # ids
        locations = _create_locations(engine, locations)

        # get all pairs of locations
        possible_routes = combinations(locations, 2)
        routes = []

        for route in possible_routes:
            routes.append(Route(route[0], route[1]))
            routes.append(Route(route[1], route[0]))

        routes = _create_routes(engine, routes)


def _create_routes(
        engine: db.Engine,
        routes: list[Route],
) -> list[Route]:

    meta_data = db.MetaData()

    table = db.Table("route", meta_data, autoload_with=engine)

    with engine.connect() as connection:

        # insert the routes
        records = [x.to_record() for x in routes]
        connection.execute(db.insert(table), records)
        connection.commit()

        response = connection.execute(db.select(table))


def _create_locations(
        engine: db.Engine,
        locations: list[Location]
) -> list[Location]:

    meta_data = db.MetaData()

    table = db.Table("location", meta_data, autoload_with=engine)

    with engine.connect() as connection:

        # insert the locations
        records = [x.to_record() for x in locations]
        connection.execute(db.insert(table), records)
        connection.commit()

        # retrieve the records with id from the database
        response = connection.execute(db.select(table))
        headers = response.keys()

        locations = []
        for row in response.fetchall():
            locations.append(dict(zip(headers, row)))

        return [Location(**x) for x in locations]


def _create_table(engine: db.Engine):

    meta_data = db.MetaData()

    db.Table(
        "location", meta_data,
        db.Column("id", db.Integer, primary_key=True, autoincrement=True),
        db.Column("name", db.String(32), nullable=False),
        db.Column("latitude", db.DECIMAL(8, 5), nullable=False),
        db.Column("longitude", db.DECIMAL(7, 5), nullable=False),
        db.UniqueConstraint("latitude", "longitude", name="unique_location"),
        db.CheckConstraint(
            "latitude >= -180 AND latitude <= 180",
            name="latitude_scope"
        ),
        db.CheckConstraint(
            "longitude >= -90 AND longitude <= 90",
            name="longitude_scope"
        ),
    )

    db.Table(
        "route", meta_data,
        db.Column("id", db.Integer, primary_key=True, autoincrement=True),
        db.Column("origin", db.ForeignKey("location.id"), nullable=False),
        db.Column("destination", db.ForeignKey("location.id"), nullable=False),
        db.UniqueConstraint("origin", "destination", name="unique_route")
    )

    db.Table(
        "route_duration", meta_data,
        db.Column("logged_time", db.DateTime),
        db.Column("route", db.ForeignKey("route.id")),
        db.Column("trip_duration", db.Integer),
        db.PrimaryKeyConstraint("logged_time", "route",
                                name="route_duration_pk"),
    )

    meta_data.create_all(engine)



"""module containing the route object"""

import datetime as dt

from sqlalchemy import Engine, Table, MetaData

from waze_logger.location import Location
from waze_logger.waze_api import get_route_duration


class Route:

    id: int
    origin: Location
    destination: Location

    def __init__(
            self,
            origin: Location,
            destination: Location,
            id: int = None,
    ):

        self.origin = origin
        self.destination = destination
        self.id = id

    def to_record(self) -> dict:
        return {
            "origin": self.origin.id,
            "destination": self.destination.id,
        }

    def duration(self) -> int:
        return get_route_duration(self.origin, self.destination)

    @staticmethod
    def from_database(engine: Engine, id: int = None) -> list["Route"]:

        table = Table("route", MetaData(), autoload_with=engine)
        statement = table.select()

        if id is not None:
            statement = statement.where(table.c.id == id)

        with engine.connect() as connection:

            response = connection.execute(statement)

            headers = response.keys()
            routes = []

            for row in response.fetchall():

                record = dict(zip(headers, row))

                # convert location id into location objects
                record["origin"] = Location.from_database(
                    engine,
                    record["origin"]
                )

                record["destination"] = Location.from_database(
                    engine,
                    record["destination"]
                )

                routes.append(Route(**record))

        if id is not None:
            return routes[0]

        return routes



""""module for a class containing location"""

from sqlalchemy import Engine, Table, MetaData


class Location:

    location_id: int
    longitude: float
    latitude: float
    name: str

    def __init__(
            self,
            latitude: float,
            longitude: float,
            name: str,
            id: int = None
    ):
        self.longitude = longitude
        self.latitude = latitude
        self.name = name
        self.id = id

    def __str__(self) -> str:
        return f"x:{self.longitude:.14f} y:{self.latitude:.14f}"

    def to_record(self) -> dict:
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "name": self.name,
        }

    @staticmethod
    def from_database(engine: Engine, id: int) -> list["Location"]:

        table = Table("location", MetaData(), autoload_with=engine)
        statement = table.select()

        if id is not None:
            statement = statement.where(table.c.id == id)

        with engine.connect() as connection:

            response = connection.execute(statement)

            headers = response.keys()
            locations = []

            for row in response.fetchall():

                record = dict(zip(headers, row))
                locations.append(Location(**record))

        if id is not None:
            return locations[0]

        return locations
